import abc
import jsonlines
import orjson
import sys
import typing as t

from .. import BaseConnector, flow
from . import request
from . import response


class Request(t.TypedDict, total=False):
    spec: request.Spec
    validate: request.Validate
    open: request.Open
    read: request.Read
    flush: request.Flush
    startCommit: request.StartCommit


class Response(t.TypedDict, total=False):
    spec: flow.Spec
    validated: response.Validated
    opened: response.Opened
    published: response.Published
    flushed: response.Flushed
    startedCommit: response.StartedCommit


class Connector(BaseConnector, metaclass=abc.ABCMeta):
    def __init__(self):
        super().__init__()

    @abc.abstractmethod
    def spec(self, spec: request.Spec) -> flow.Spec:
        raise NotImplemented

    @abc.abstractmethod
    def validate(self, validate: request.Validate) -> response.Validated:
        raise NotImplemented

    @abc.abstractmethod
    def open(self, open: request.Open) -> None:
        raise NotImplemented

    @abc.abstractmethod
    def read(
        self, read: request.Read, emit: t.Callable[[response.Published], None]
    ) -> None:
        raise NotImplemented

    def flush(
        self, flush: request.Flush, emit: t.Callable[[response.Published], None]
    ) -> response.Flushed:
        return response.Flushed()

    def startCommit(self, start_commit: request.StartCommit) -> response.StartedCommit:
        return response.StartedCommit(
            state=flow.ConnectorState(updated={}, mergePatch=False)
        )

    def _run(self) -> None:
        requests = jsonlines.Reader(sys.stdin, loads=orjson.loads)
        request: Request

        for request in requests:
            if (spec := request.get("spec")) is not None:
                spec_response = self.spec(spec)
                spec_response["protocol"] = 3032023
                self._emit(Response(spec=spec_response))
            elif (validate := request.get("validate")) is not None:
                self._emit(Response(validated=self.validate(validate)))
            elif (open := request.get("open")) is not None:
                self.open(open)
                self._emit(Response(opened=response.Opened()))
                break

        def _emit_published(published: response.Published) -> None:
            self._emit(Response(published=published))

        for request in requests:
            if (read := request.get("read")) is not None:
                self.read(read, _emit_published)
            elif (flush := request.get("flush")) is not None:
                self._emit(Response(flushed=self.flush(flush, _emit_published)))
            elif (start := request.get("startCommit")) is not None:
                self._emit(Response(startedCommit=self.startCommit(start)))
            else:
                raise RuntimeError("malformed request", request)



import typing as t

ConnectorType = t.Literal["IMAGE", "LOCAL"]

class CollectionSpec(t.TypedDict):
    name: str
    writeSchema: dict
    readSchema: dict
    key: t.List[str]

class CaptureBinding(t.TypedDict):
    collection: CollectionSpec
    resourceConfig: dict
    resourcePath: t.List[str]
    backfill: int
    stateKey: str

class CaptureSpec(t.TypedDict):
    name: str
    connectorType: ConnectorType
    config: dict
    intervalSeconds: int
    bindings: t.List[CaptureBinding]

class RangeSpec(t.TypedDict):
    keyBegin: t.Optional[int]
    keyEnd: int
    rClockBegin: t.Optional[int]
    rClockEnd: int

class UUIDParts(t.TypedDict):
    node: str
    clock: str

class CheckpointSource(t.TypedDict):
    readThrough: str
    producers: t.List[t.Any]

class Checkpoint(t.TypedDict):
    sources: dict[str, CheckpointSource]
    ackIntents: dict[str, str]

class OAuth2(t.TypedDict):
    provider: str
    accessTokenBody: str
    authUrlTemplate: str
    accessTokenHeaders: t.Dict[str, str]
    accessTokenResponseMap: t.Dict[str, str]
    accessTokenUrlTemplate: str

class Spec(t.TypedDict):
    protocol: t.NotRequired[int]
    configSchema: dict
    resourceConfigSchema: dict
    documentationUrl: str
    oauth2: t.NotRequired[OAuth2]
    resourcePathPointers: t.List[str]

class ConnectorState(t.TypedDict):
    updated: dict
    mergePatch: bool



from .. import flow
import typing as t

class DiscoveredBinding(t.TypedDict):
    recommendedName: str
    resourceConfig: dict
    documentSchema: dict
    key: t.List[str]
    disable: t.NotRequired[bool]

class Discovered(t.TypedDict):
    bindings: t.List[DiscoveredBinding]

class ValidatedBinding(t.TypedDict):
    resourcePath: t.List[str]

class Validated(t.TypedDict):
    bindings: t.List[ValidatedBinding]

class Applied(t.TypedDict, total=False):
    actionDescription: str

class Opened(t.TypedDict):
    explicitAcknowledgements: bool

class Captured(t.TypedDict):
    binding: int
    doc: dict

class Checkpoint(t.TypedDict):
    state: flow.ConnectorState



import abc
import orjson
import os
import sys
import typing as t
from .logger import init_logger

logger = init_logger()


class BaseConnector(metaclass=abc.ABCMeta):
    @abc.abstractmethod
    def _run(self) -> None:
        raise NotImplemented

    def _emit(self, response: t.Any) -> None:
        # TODO(johnny): Make this buffered?
        os.write(
            1,
            orjson.dumps(
                response,
                # Map non-string dict keys into str
                option=orjson.OPT_NON_STR_KEYS
                # Prefer 'Z' over '+00:00'.
                | orjson.OPT_UTC_Z
                # Add a newline for newline JSON.
                | orjson.OPT_APPEND_NEWLINE,
                # Map unhandled JSON types through __str__() if available, or repr().
                default=str,
            ),
        )

    def main(self) -> None:
        try:
            self._run()

        except ValidateError as exc:
            logger.exception(exc.message)
            sys.exit(-1)

        except Exception as exc:
            msg, args = type(exc).__name__, exc.args
            if len(args) != 0:
                msg = f"{msg}: {args[0]}"

            logger.exception(msg, args)
            sys.exit(-1)


class ValidateError(Exception):
    def __init__(self, message: str) -> None:
        super().__init__(message)
        self.message = message



import logging
import os
import typing as t
from airbyte_cdk.sources.source import Source
from dataclasses import asdict, dataclass
from airbyte_protocol.models import (
    SyncMode,
    ConfiguredAirbyteCatalog,
    ConfiguredAirbyteStream,
    AirbyteStateMessage,
    AirbyteStreamState,
    AirbyteGlobalState,
    AirbyteStateType,
    AirbyteStream,
    Status,
    Level as LogLevel,
)

from .capture import Connector, request, response, Response
from . import flow, ValidateError, logger
from .logger import init_logger

logger = init_logger()

# `logger` has name "flow", and we thread it through the Airbyte APIs,
# but connectors may still get their own "airbyte" logger.
# Patch it to use the same log level as the "flow" Logger.
logging.getLogger("airbyte").setLevel(logger.level)

DOCS_URL = os.getenv("DOCS_URL")


"""
Airbyte doesn't appear to reduce state like we do. As a result,
the Airbyte CDK is expecting its state input to look like a list of
AirbyteStateMessages. Since we _do_ reduce state, what we do here is
keep track of the latest AirbyteStateMessage for each stream, and then
just give Airbyte a list of all of the latest state messages on boot.
"""
@dataclass
class State:
    stream_state: t.Optional[t.Dict[str, AirbyteStateMessage]] = None
    global_state: t.Optional[AirbyteStateMessage] = None

    def handle_message(self, msg: AirbyteStateMessage):
        if msg.type == AirbyteStateType.STREAM:
            if msg.stream == None:
                raise Exception(
                    "Got a STREAM-specific state message with no stream-specific state"
                )
            if not self.stream_state:
                self.stream_state = {}
            self.stream_state[
                f"{msg.stream.stream_descriptor.name}-{msg.stream.stream_descriptor.namespace}"
            ] = msg
        elif msg.type == AirbyteStateType.GLOBAL:
            self.global_state = msg
        elif msg.type == AirbyteStateType.LEGACY or msg.type is None:
            raise Exception("Airbyte LEGACY state is not supported")

    def to_airbyte_input(self):
        if self.global_state:
            return [self.global_state]
        elif self.stream_state:
            return [msg for k, msg in self.stream_state.items()]


class CaptureShim(Connector):
    delegate: Source
    usesSchemaInference: bool

    def __init__(
        self, delegate: Source, oauth2: flow.OAuth2 | None, usesSchemaInference=True
    ):
        super().__init__()
        self.delegate = delegate
        self.oauth2 = oauth2
        self.usesSchemaInference = usesSchemaInference

    def spec(self, _: request.Spec) -> flow.Spec:
        spec = self.delegate.spec(logger)

        out: flow.Spec = {
            "documentationUrl": f"{DOCS_URL if DOCS_URL else spec.documentationUrl}",
            "configSchema": spec.connectionSpecification,
            "resourceConfigSchema": resource_config_schema,
            "resourcePathPointers": ["/namespace", "/stream"],
        }

        # TODO(johnny): Can we map spec.advanced_auth into flow.OAuth2 ?
        if self.oauth2:
            out["oauth2"] = self.oauth2

        return out

    def discover(self, discover: request.Discover) -> response.Discovered:
        catalog = self.delegate.discover(logger, discover["config"])

        bindings: t.List[response.DiscoveredBinding] = []

        for stream in catalog.streams:
            if SyncMode.incremental in stream.supported_sync_modes:
                sync_mode = SyncMode.incremental
            elif SyncMode.full_refresh in stream.supported_sync_modes:
                sync_mode = SyncMode.full_refresh
            else:
                raise RuntimeError("invalid sync modes", stream.supported_sync_modes)

            config: StreamResourceConfig = {
                "stream": stream.name,
                "syncMode": sync_mode,
            }
            if stream.namespace:
                config["namespace"] = stream.namespace

            json_schema = stream.json_schema

            if stream.source_defined_primary_key:
                # Map array of array of property names into an array of JSON pointers.
                key = [
                    "/"
                    + "/".join(
                        p.replace("~", "~0").replace("/", "~1") for p in component
                    )
                    for component in stream.source_defined_primary_key
                ]
            elif sync_mode == SyncMode.full_refresh:
                # Synthesize a key based on the record's order within each stream refresh.
                key = ["/_meta/row_id"]

                # Extend schema with /_meta/row_id.
                meta = stream.json_schema.setdefault("properties", {}).setdefault(
                    "_meta", {"type": "object"}
                )
                meta.setdefault("properties", {})["row_id"] = {"type": "integer"}
                meta.setdefault("required", []).append("row_id")
                json_schema = stream.json_schema
            else:
                raise RuntimeError(
                    "incremental stream is missing a source-defined primary key",
                    stream.name,
                )

            if self.usesSchemaInference:
                json_schema["x-infer-schema"] = True

            bindings.append(
                response.DiscoveredBinding(
                    recommendedName=stream.name,
                    documentSchema=json_schema,
                    resourceConfig=t.cast(dict, config),
                    key=key,
                )
            )

        return response.Discovered(bindings=bindings)

    def validate(self, validate: request.Validate) -> response.Validated:
        result = self.delegate.check(logger, validate["config"])
        if result.status != Status.SUCCEEDED:
            raise ValidateError(f"{result.message}")

        bindings: t.List[response.ValidatedBinding] = []

        for b in validate["bindings"]:
            config = t.cast(StreamResourceConfig, b["resourceConfig"])

            if ns := config.get("namespace"):
                resource_path = [ns, config["stream"]]
            else:
                resource_path = [config["stream"]]

            bindings.append(response.ValidatedBinding(resourcePath=resource_path))

        return response.Validated(bindings=bindings)

    def open(self, open: request.Open, emit: t.Callable[[Response], None]) -> None:
        streams: t.List[ConfiguredAirbyteStream] = []

        # Key: (namespace, stream).
        # Value: [binding index, next row_id].
        index: dict[t.Tuple[str | None, str], t.List[int]] = {}

        for i, binding in enumerate(open["capture"]["bindings"]):
            rc = t.cast(StreamResourceConfig, binding["resourceConfig"])

            # TODO(johnny): Re-hydrate next row_id from open["state"].
            index[(rc.get("namespace"), rc["stream"])] = [i, 1]

            streams.append(
                ConfiguredAirbyteStream(
                    stream=AirbyteStream(
                        json_schema=binding["collection"]["writeSchema"],
                        name=rc["stream"],
                        namespace=rc.get("namespace"),
                        supported_sync_modes=[rc["syncMode"]],
                    ),
                    cursor_field=rc.get("cursorField"),
                    destination_sync_mode="append",
                    sync_mode=rc["syncMode"],
                )
            )

        catalog = ConfiguredAirbyteCatalog(streams=streams)
        config = open["capture"]["config"]
        state = State(**(open["state"] or {}))

        emit(Response(opened=response.Opened(explicitAcknowledgements=False)))

        for message in self.delegate.read(
            logger, config, catalog, state.to_airbyte_input()
        ):
            if record := message.record:
                entry = index[(record.namespace, record.stream)]
                record.data.setdefault("_meta", {})["row_id"] = entry[1]
                entry[1] += 1

                emit(
                    Response(
                        captured=response.Captured(binding=entry[0], doc=record.data)
                    )
                )

            elif state_msg := message.state:
                state_msg = t.cast(AirbyteStateMessage, state_msg)
                logger.info(f"Got a state message: {str(state_msg)}")
                # TODO(johnny): mix in next_row_id
                state.handle_message(state_msg)
                emit(
                    Response(
                        checkpoint=response.Checkpoint(
                            state=flow.ConnectorState(
                                updated=asdict(state), mergePatch=False
                            )
                        )
                    )
                )

            elif trace := message.trace:
                if error := trace.error:
                    logger.error(
                        error.message,
                        extra={
                            "internal_message": error.internal_message,
                            "stack_trace": error.stack_trace,
                        },
                    )
                elif estimate := trace.estimate:
                    logger.info(
                        "progress estimate",
                        extra={
                            "estimate": {
                                "stream": estimate.name,
                                "namespace": estimate.namespace,
                                "type": estimate.type,
                                "row_estimate": estimate.row_estimate,
                                "byte_estimate": estimate.byte_estimate,
                            }
                        },
                    )
                elif status := trace.stream_status:
                    logger.info(
                        "stream status",
                        extra={
                            "status": {
                                "stream": status.stream_descriptor.name,
                                "namespace": status.stream_descriptor.namespace,
                                "status": status.status,
                            }
                        },
                    )
                else:
                    raise RuntimeError(f"unexpected trace type {trace.type}")

            elif log := message.log:
                if log.level == LogLevel.DEBUG:
                    level = logging.DEBUG
                elif log.level == LogLevel.INFO:
                    level = logging.INFO
                elif log.level == LogLevel.WARN:
                    level = logging.WARNING
                else:
                    level = logging.ERROR

                logger.log(level, log.message, log.stack_trace)

            else:
                raise RuntimeError("unexpected AirbyteMessage", message)
        # Emit a final state message
        emit(
            Response(
                checkpoint=response.Checkpoint(
                    state=flow.ConnectorState(updated=asdict(state), mergePatch=False)
                )
            )
        )


class StreamResourceConfig(t.TypedDict):
    stream: str
    syncMode: SyncMode
    namespace: t.NotRequired[str]
    cursorField: t.NotRequired[t.List[str]]


resource_config_schema = {
    "type": "object",
    "properties": {
        "stream": {"type": "string"},
        "syncMode": {"type": "string", "enum": ["incremental", "full_refresh"]},
        "namespace": {"type": "string"},
        "cursorField": {
            "type": "array",
            "items": {"type": "string"},
        },
    },
    "required": ["stream", "syncMode"],
}



from .. import flow
import typing as t

class ValidatedTransform(t.TypedDict):
    readOnly: bool

class Validated(t.TypedDict):
    transforms: t.List[ValidatedTransform]
    generatedFiles: dict[str, str]

class Opened(t.TypedDict):
    pass

class Published(t.TypedDict):
    doc: dict

class Flushed(t.TypedDict):
    pass 

class StartedCommit(t.TypedDict):
    state: flow.ConnectorState


import os
import sys
import singer_sdk._singerlib as singer
import singer_sdk.metrics
from singer_sdk.streams import Stream
from singer_sdk.typing import (
    DateTimeType,
    IntegerType,
    ObjectType,
    PropertiesList,
    Property,
)
import typing as t

from .capture import Connector, request, response, Response
from . import flow, ValidateError, init_logger

logger = init_logger()

def write_message_panics(message: singer.Message) -> None:
    raise RuntimeError("unexpected call to singer.write_message", message.to_dict())

# Disable the _setup_logging routine, as it otherwise clobbers our logging setup.
singer_sdk.metrics._setup_logging = lambda _config: None
# The singer-sdk uses LOGLEVEL to set up loggers instead of LOG_LEVEL.
os.environ.setdefault("LOGLEVEL", os.environ.get("LOG_LEVEL", "INFO").upper())
# CaptureShim instruments singer.write_message as-needed to capture message callbacks.
# We don't expect it to ever be called outside of those specific contexts.
singer.write_message = write_message_panics

DOCS_URL = os.getenv("DOCS_URL")

# Replication modes from singer_sdk/streams/core.py
REPLICATION_FULL_TABLE = "FULL_TABLE"
REPLICATION_INCREMENTAL = "INCREMENTAL"

class Config(t.TypedDict):
    pass

class State(t.TypedDict):
    pass


class DelegateFactoryCallable(t.Protocol):
    def __call__(self, config: Config, catalog: singer.Catalog | None = None, state: State | None = None) -> singer_sdk.Tap:
        ...

class CaptureShim(Connector):
    config_schema: dict
    delegate_factory: DelegateFactoryCallable
    usesSchemaInference: bool
    oauth2: flow.OAuth2 | None

    def __init__(
        self,
        config_schema: dict,
        delegate_factory: DelegateFactoryCallable,
        usesSchemaInference = True,
        oauth2: flow.OAuth2 | None = None
    ):
        super().__init__()
        self.config_schema = config_schema
        self.delegate_factory = delegate_factory
        self.usesSchemaInference = usesSchemaInference
        self.oauth2 = oauth2

    def spec(self, _: request.Spec) -> flow.Spec:
        out = flow.Spec(
            documentationUrl=DOCS_URL if DOCS_URL else "https://docs.estuary.dev",
            configSchema=self.config_schema,
            resourceConfigSchema=resource_config_schema,
            resourcePathPointers=["/stream"],
        )

        if self.oauth2:
           out["oauth2"] = self.oauth2

        return out

    def discover(self, discover: request.Discover) -> response.Discovered:
        config = t.cast(Config, discover["config"])
        delegate = self.delegate_factory(config=config)

        bindings: t.List[response.DiscoveredBinding] = []

        # Use stack-based iteration to enumerate all child streams.
        stack = [s for s in delegate.discover_streams()]

        while stack:
            stream: singer_sdk.Stream = stack.pop(0)
            stack.extend(stream.child_streams)

            entry: singer.CatalogEntry = stream._singer_catalog_entry

            if entry.replication_method == REPLICATION_INCREMENTAL and not entry.replication_key:
                raise RuntimeError(f"Stream {stream.name} must have a replication key in order to use {REPLICATION_INCREMENTAL} replication.")
            
            # According to the Singer spec, key_properties are top-level
            # properties of the document. We extend this interpretation to
            # also allow for arbitrary JSON pointers.
            if entry.key_properties:
                key = [
                    p
                    if p.startswith("/")
                    else "/" + p.replace("~", "~0").replace("/", "~1")
                    for p in entry.key_properties
                ]
            else:
                logger.info(f"Stream {stream.name} does not have a primary key, using '/_meta/row_id' as the key.") 
                if entry.replication_method == REPLICATION_INCREMENTAL:
                    logger.warn(f"Stream {stream.name} is using {REPLICATION_INCREMENTAL} replication without a primary key. If this is not an append-only data source, updates to existing documents will be captured as inserts with new keys. Switch to {REPLICATION_FULL_TABLE} to fix this.")
                
                key = ["/_meta/row_id"]
                row_id_type = Property("row_id", IntegerType, required=True)

                if entry.schema.properties is not None:
                    if not "_meta" in entry.schema.properties:
                        entry.schema.properties["_meta"] = ObjectType(row_id_type)
                        if not entry.schema.required:
                            entry.schema.required = []
                        if not "_meta" in entry.schema.required:
                            entry.schema.required.append("_meta")
                    else:
                        meta = entry.schema.properties["_meta"]
                        
                        # Just in case _meta was somehow not required
                        meta.optional = False
                        props: ObjectType = meta.wrapped
                        props.wrapped["row_id"] = row_id_type
                else:
                    # We should never get here, this is just to appease type checking
                    raise RuntimeError(f"Something unexpected happened: schema for stream {stream.name} does not have a properties field.", vars(entry.schema))

            for bc, meta in entry.metadata.items():
                if meta.inclusion == "available":
                    meta.selected = True

            json_schema = entry.schema.to_dict()
            resourceConfig = entry.to_dict()

            if self.usesSchemaInference:
                json_schema["x-infer-schema"] = True

            bindings.append(
                response.DiscoveredBinding(
                    recommendedName=entry.tap_stream_id,
                    documentSchema=json_schema,
                    resourceConfig=resourceConfig,
                    key=key,
                )
            )

        return response.Discovered(bindings=bindings)

    def validate(self, validate: request.Validate) -> response.Validated:
        catalog = singer.Catalog()
        config = t.cast(Config, validate["config"])
        bindings: t.List[response.ValidatedBinding] = []

        for binding in validate["bindings"]:
            entry = singer.CatalogEntry.from_dict(binding["resourceConfig"])
            entry.schema = singer.Schema.from_dict(binding["collection"]["writeSchema"])
            catalog[entry.tap_stream_id] = entry

            # It looks[1] like it's incorrect to "upgrade" a stream that does not
            # have a replication or primary key to use `INCREMENTAL` mode, but it should
            # be possible to "downgrade" an INCREMENTAL stream to FULL_TABLE
            # [1]: https://github.com/meltano/sdk/blob/main/singer_sdk/streams/core.py#L751-L756

            if entry.replication_method == "INCREMENTAL" and entry.replication_key is None:
                raise ValidateError(f"{entry.stream} does not support {entry.replication_method} replication because it does not have a replication key.")

            bindings.append(
                response.ValidatedBinding(resourcePath=[entry.tap_stream_id])
            )

        delegate = self.delegate_factory(config=config, catalog=catalog)

        def _on_message(message: singer.Message) -> None:
            logger.debug("connection test message", message)

        singer.write_message = _on_message
        # run_connection_test() attempts to read from streams that are not selected
        # which we don't want -- we only want to validate that streams we care about work,
        # not that all possible streams work.
        _ = delegate.run_sync_dry_run(
            dry_run_record_limit=1,
            streams=[stream for stream in delegate.streams.values() if stream.selected],
        )
        singer.write_message = write_message_panics

        return response.Validated(bindings=bindings)

    def open(self, open: request.Open, emit: t.Callable[[Response], None]) -> None:
        catalog = singer.Catalog()
        config = t.cast(Config, open["capture"]["config"])
        state = t.cast(State, open["state"])

        # Key: tap_stream_id.
        # Value: [binding index, next row_id].
        index: dict[str, t.List[int]] = state.get("rowIds", {}) # type: ignore

        for i, binding in enumerate(open["capture"]["bindings"]):
            entry = singer.CatalogEntry.from_dict(binding["resourceConfig"])
            entry.schema = singer.Schema.from_dict(binding["collection"]["writeSchema"])

            # TODO(johnny): Re-hydrate next row_id from open["state"].
            index[entry.tap_stream_id] = [i, 1]
            catalog[entry.tap_stream_id] = entry

        delegate: singer_sdk.Tap = self.delegate_factory(config=config, catalog=catalog, state=state.get("bindingStateV1",{})) # type: ignore
        emit(Response(opened=response.Opened(explicitAcknowledgements=False)))

        def _on_message(message: singer.Message):
            if message.type == singer.SingerMessageType.SCHEMA:
                return  # Ignored.
            elif message.type == singer.SingerMessageType.RECORD:
                record = t.cast(singer.RecordMessage, message)
                entry = index[record.stream]

                record.record.setdefault("_meta", {})["row_id"] = entry[1]
                entry[1] += 1

                emit(
                    Response(
                        captured=response.Captured(binding=entry[0], doc=record.record)
                    )
                )
            elif message.type == singer.SingerMessageType.STATE:
                state = t.cast(singer.StateMessage, message)
                logger.debug("stream state", state)
                # The Meltano SDK manages any stream-specific state for us.
                # See their docs on this here: https://sdk.meltano.com/en/latest/implementation/state.html
                emit(
                    Response(
                        checkpoint=response.Checkpoint(
                            state=flow.ConnectorState(
                                updated={
                                    # TODO: unwrap/re-wrap this state using resourcepathpointers
                                    "bindingStateV1": state.to_dict(),
                                    "rowIds": index
                                },
                                mergePatch=False
                            )
                        )
                    )
                )
            else:
                raise RuntimeError("unexpected singer_sdk Message", message)

        singer.write_message = _on_message
        self.sync_streams(delegate)
        singer.write_message = write_message_panics

    # Lifted from singer-sdk for the moment, but pulling this in-tree will
    # allow us to do fancy things like driving streams in parallel
    # instead of the current implementation which drives them sequentially.
    def sync_streams(self, delegate: singer_sdk.Tap): 
        delegate._reset_state_progress_markers()
        delegate._set_compatible_replication_methods()
        singer.write_message(singer.StateMessage(value=delegate.state))

        stream: Stream
        for stream in delegate.streams.values():
            if not stream.selected and not stream.has_selected_descendents:
                delegate.logger.info("Skipping deselected stream '%s'.", stream.name)
                continue

            if stream.parent_stream_type:
                delegate.logger.debug(
                    "Child stream '%s' is expected to be called "
                    "by parent stream '%s'. "
                    "Skipping direct invocation.",
                    type(stream).__name__,
                    stream.parent_stream_type.__name__,
                )
                continue

            delegate.logger.info("Syncing stream '%s'.", stream.name)
            stream.sync()
            # This is where we tell the Meltano SDK to do any final state emits if neccesary
            stream.finalize_state_progress_markers()

        # this second loop is needed for all streams to print out their costs
        # including child streams which are otherwise skipped in the loop above
        for stream in delegate.streams.values():
            stream.log_sync_costs()


resource_config_schema = {
    "type": "object",
    "properties": {
        "stream": {"type": "string"},
        "replication_method": {"type": "string", "enum": [REPLICATION_INCREMENTAL, REPLICATION_FULL_TABLE]},
        "replication_key": {"type": "string"}
    },
    "required": ["stream", "replication_method"]
}



from .. import flow
import typing as t

ConnectorType = t.Literal["IMAGE", "LOCAL", "SQLITE", "TYPESCRIPT"]

class Spec(t.TypedDict):
    connectorType: ConnectorType
    config: dict

class ValidateTransform(t.TypedDict):
    name: str
    collection: flow.CollectionSpec
    shuffleLambdaConfig: t.Optional[dict]
    lambdaConfig: dict

class Validate(t.TypedDict):
    connectorType: ConnectorType
    config: dict
    collection: flow.CollectionSpec
    transforms: t.List[ValidateTransform]
    shuffleKeyTypes: t.List[str]
    projectRoot: str
    importMap: dict[str, str]

class Open(t.TypedDict):
    collection: flow.CollectionSpec
    version: str
    range: flow.RangeSpec
    state: dict

class Shuffle(t.TypedDict):
    key: str
    packed: str
    hash: int

class Read(t.TypedDict):
    transform: int
    uuid: flow.UUIDParts
    shuffle: Shuffle
    doc: dict

class Flush(t.TypedDict):
    pass # Always empty.

class StartCommit(t.TypedDict):
    runtimeCheckpoint: flow.Checkpoint

class Reset(t.TypedDict):
    pass # Always empty.



from queue import Queue
from threading import Thread
import abc
import jsonlines
import orjson
import sys
import typing as t

from .. import BaseConnector, flow
from . import request
from . import response


class Request(t.TypedDict, total=False):
    spec: request.Spec
    discover: request.Discover
    validate: request.Validate
    apply: request.Apply
    open: request.Open
    acknowledge: request.Acknowledge


class Response(t.TypedDict, total=False):
    spec: flow.Spec
    discovered: response.Discovered
    validated: response.Validated
    applied: response.Applied
    opened: response.Opened
    captured: response.Captured
    checkpoint: response.Checkpoint


class Connector(BaseConnector, metaclass=abc.ABCMeta):
    def __init__(self):
        super().__init__()

    @abc.abstractmethod
    def spec(self, spec: request.Spec) -> flow.Spec:
        raise NotImplementedError()

    @abc.abstractmethod
    def discover(self, discover: request.Discover) -> response.Discovered:
        raise NotImplementedError()

    @abc.abstractmethod
    def validate(self, validate: request.Validate) -> response.Validated:
        raise NotImplementedError()

    def apply(self, apply: request.Apply) -> response.Applied:
        return response.Applied()  # No-op.

    @abc.abstractmethod
    def open(self, open: request.Open, emit: t.Callable[[Response], None]) -> None:
        raise NotImplementedError()

    def acknowledge(self, acknowledge: request.Acknowledge) -> None:
        pass  # No-op.

    def _run(self) -> None:
        # Start a read loop in a thread which dispatches to `requests`.
        requests: Queue[Request | None] = Queue(maxsize=8)
        Thread(
            target=_read_loop, name="read loop", args=(requests,), daemon=True
        ).start()

        while True:
            if not (request := requests.get()):
                return  # EOF.
            elif (spec := request.get("spec")) is not None:
                spec_response = self.spec(spec)
                spec_response["protocol"] = 3032023
                self._emit(Response(spec=spec_response))
            elif (discover := request.get("discover")) is not None:
                self._emit(Response(discovered=self.discover(discover)))
            elif (validate := request.get("validate")) is not None:
                self._emit(Response(validated=self.validate(validate)))
            elif (apply := request.get("apply")) is not None:
                self._emit(Response(applied=self.apply(apply)))
            elif (open := request.get("open")) is not None:
                break
            else:
                raise RuntimeError("malformed request", request)

        def _emit_response(response: Response) -> None:
            self._emit(response)

            # Queue rarely has items, so optimize for the common "empty" case.
            # Otherwise non-blocking get() throws Empty, and exceptions are expensive.
            # This is racy and can false-negative but can never false-positive,
            # since we're the only queue consumer.
            if requests.qsize() == 0:
                return

            if not (next := requests.get()):
                return  # EOF
            elif (acknowledge := next.get("acknowledge")) is not None:
                self.acknowledge(acknowledge)
            else:
                raise RuntimeError("malformed request", request)

        self.open(open, _emit_response)


def _read_loop(queue: Queue[Request | None]):
    request: Request
    for request in jsonlines.Reader(sys.stdin, loads=orjson.loads):
        queue.put(request)

    queue.put(None)  # Signal EOF.



import logging.config
import orjson
import os


class LogFormatter(logging.Formatter):
    # Keys which are present in all LogRecord instances.
    # We use this set to identify _novel_ keys which should be included as structured, logged fields.
    LOGGING_RECORD_KEYS = logging.LogRecord(
        "", 0, "", 0, None, None, None
    ).__dict__.keys()

    def format(self, record: logging.LogRecord) -> str:
        # Attach any extra keywords which are not ordinarily in a LogRecord as fields.
        fields = {
            k: getattr(record, k)
            for k in record.__dict__.keys()
            if hasattr(record, k) and k not in self.LOGGING_RECORD_KEYS
        }
        if record.args:
            fields["args"] = record.args

        fields["source"] = record.name
        fields["file"] = f"{record.pathname}:{record.lineno}"

        # Attach any included stack traces.
        if record.exc_info:
            fields["traceback"] = self.formatException(record.exc_info).splitlines()
        elif record.stack_info:
            fields["stack"] = self.formatStack(record.stack_info).splitlines()

        return str(
            orjson.dumps(
                {
                    "level": record.levelname,
                    "msg": record.msg,
                    "fields": fields,
                },
                # Map non-string dict keys into str, and prefer 'Z' over '+00:00'.
                option=orjson.OPT_NON_STR_KEYS | orjson.OPT_UTC_Z,
                # Map unhandled JSON types through __str__() if available, or repr().
                default=str,
            ),
            encoding="utf-8",
            errors="ignore",
        )


def init_logger():
    LOGGING_CONFIG = {
        "version": 1,
        "disable_existing_loggers": False,
        "formatters": {
            "flow": {
                "()": LogFormatter,
                "format": "",
            },
        },
        "handlers": {
            "console": {
                "class": "logging.StreamHandler",
                "stream": "ext://sys.stderr",
                "formatter": "flow",
            },
        },
        "root": {
            "handlers": ["console"],
        },
    }

    logging.config.dictConfig(LOGGING_CONFIG)

    logger = logging.getLogger("flow")
    logger.setLevel(os.environ.get("LOG_LEVEL", "INFO").upper())

    return logger



from .. import flow
import typing as t

class Spec(t.TypedDict):
    connectorType: flow.ConnectorType
    config: dict

class Discover(t.TypedDict):
    connectorType: flow.ConnectorType
    config: dict

class ValidateBinding(t.TypedDict):
    collection: flow.CollectionSpec
    resourceConfig: dict

class Validate(t.TypedDict):
    name: str
    connectorType: flow.ConnectorType
    config: dict
    bindings: t.List[ValidateBinding]

class Apply(t.TypedDict):
    capture: flow.CaptureSpec
    version: str
    dryRun: bool

class Open(t.TypedDict):
    capture: flow.CaptureSpec
    version: str
    range: flow.RangeSpec
    state: dict

class Acknowledge(t.TypedDict):
    checkpoints: int